#include "datablock_logging.h"

extern "C" {

const char * BLOCK_LOG_READ = "READ-OK";
const char * BLOCK_LOG_WRITE = "WRITE-OK";
const char * BLOCK_LOG_READ_FAIL = "READ-FAIL";
const char * BLOCK_LOG_WRITE_FAIL = "WRITE-FAIL";
const char * BLOCK_LOG_READ_DEFAULT = "READ-DEFAULT";
const char * BLOCK_LOG_REPLACE = "REPLACE-OK";
const char * BLOCK_LOG_REPLACE_FAIL = "REPLACE-FAIL";
const char * BLOCK_LOG_CLEAR = "CLEAR";
const char * BLOCK_LOG_DELETE = "DELETE";
const char * BLOCK_LOG_START_MODULE = "MODULE-START";
const char * BLOCK_LOG_COPY = "COPY";

}