
! This module is auto-generated from the file cosmosis/datablock/section_names.txt.
! Edit that and then re-make to add your own pre-defined section names.

module cosmosis_section_names


    implicit none
    character(*), parameter :: likelihoods_section = "likelihoods"
    character(*), parameter :: data_vector_section = "data_vector"
    character(*), parameter :: cosmological_parameters_section = "cosmological_parameters"
    character(*), parameter :: halo_model_parameters_section = "halo_model_parameters"
    character(*), parameter :: baryon_parameters_section = "baryon_parameters"
    character(*), parameter :: shear_calibration_parameters_section = "shear_calibration_parameters"
    character(*), parameter :: number_density_params_section = "number_density_params"
    character(*), parameter :: post_friedmann_parameters_section = "post_friedmann_parameters"
    character(*), parameter :: intrinsic_alignment_parameters_section = "intrinsic_alignment_parameters"
    character(*), parameter :: intrinsic_power_section = "intrinsic_power"
    character(*), parameter :: galaxy_intrinsic_power_section = "galaxy_intrinsic_power"
    character(*), parameter :: matter_intrinsic_power_section = "matter_intrinsic_power"
    character(*), parameter :: wl_number_density_section = "wl_number_density"
    character(*), parameter :: matter_power_nl_section = "matter_power_nl"
    character(*), parameter :: matter_power_lin_section = "matter_power_lin"
    character(*), parameter :: matter_power_lin_cdm_baryon_section = "matter_power_lin_cdm_baryon"
    character(*), parameter :: matter_galaxy_power_section = "matter_galaxy_power"
    character(*), parameter :: galaxy_power_section = "galaxy_power"
    character(*), parameter :: shear_xi_section = "shear_xi"
    character(*), parameter :: shear_cl_section = "shear_cl"
    character(*), parameter :: shear_cl_gg_section = "shear_cl_gg"
    character(*), parameter :: shear_cl_gi_section = "shear_cl_gi"
    character(*), parameter :: shear_cl_ii_section = "shear_cl_ii"
    character(*), parameter :: galaxy_cl_section = "galaxy_cl"
    character(*), parameter :: cmb_cl_section = "cmb_cl"
    character(*), parameter :: lss_autocorrelation_section = "lss_autocorrelation"
    character(*), parameter :: gal_matter_power_lin_section = "gal_matter_power_lin"
    character(*), parameter :: linear_cdm_transfer_section = "linear_cdm_transfer"
    character(*), parameter :: sigma_r_lin_section = "sigma_r_lin"
    character(*), parameter :: sigma_r_nl_section = "sigma_r_nl"
    character(*), parameter :: galaxy_xi_section = "galaxy_xi"
    character(*), parameter :: galaxy_shear_xi_section = "galaxy_shear_xi"
    character(*), parameter :: galaxy_intrinsic_xi_section = "galaxy_intrinsic_xi"
    character(*), parameter :: planck_section = "planck"
    character(*), parameter :: intrinsic_alignment_field_section = "intrinsic_alignment_field"
    character(*), parameter :: fgas_parameters_section = "fgas_parameters"
    character(*), parameter :: shear_calibration_section = "shear_calibration"
    character(*), parameter :: supernova_params_section = "supernova_params"
    character(*), parameter :: bias_field_section = "bias_field"
    character(*), parameter :: galaxy_luminosity_function_section = "galaxy_luminosity_function"
    character(*), parameter :: distances_section = "distances"
    character(*), parameter :: de_equation_of_state_section = "de_equation_of_state"
    character(*), parameter :: growth_parameters_section = "growth_parameters"
    character(*), parameter :: test_parameters_section = "test_parameters"
    character(*), parameter :: mass_function_section = "mass_function"
    character(*), parameter :: evs_section = "evs"
    character(*), parameter :: clusters_section = "clusters"

end module

