
# This module is auto-generated from the file cosmosis/datablock/section_names.txt.
# Edit that and then re-make to add your own pre-defined section names.

likelihoods = "likelihoods"
data_vector = "data_vector"
cosmological_parameters = "cosmological_parameters"
halo_model_parameters = "halo_model_parameters"
baryon_parameters = "baryon_parameters"
shear_calibration_parameters = "shear_calibration_parameters"
number_density_params = "number_density_params"
post_friedmann_parameters = "post_friedmann_parameters"
intrinsic_alignment_parameters = "intrinsic_alignment_parameters"
intrinsic_power = "intrinsic_power"
galaxy_intrinsic_power = "galaxy_intrinsic_power"
matter_intrinsic_power = "matter_intrinsic_power"
wl_number_density = "wl_number_density"
matter_power_nl = "matter_power_nl"
matter_power_lin = "matter_power_lin"
matter_power_lin_cdm_baryon = "matter_power_lin_cdm_baryon"
matter_galaxy_power = "matter_galaxy_power"
galaxy_power = "galaxy_power"
shear_xi = "shear_xi"
shear_cl = "shear_cl"
shear_cl_gg = "shear_cl_gg"
shear_cl_gi = "shear_cl_gi"
shear_cl_ii = "shear_cl_ii"
galaxy_cl = "galaxy_cl"
cmb_cl = "cmb_cl"
lss_autocorrelation = "lss_autocorrelation"
gal_matter_power_lin = "gal_matter_power_lin"
linear_cdm_transfer = "linear_cdm_transfer"
sigma_r_lin = "sigma_r_lin"
sigma_r_nl = "sigma_r_nl"
galaxy_xi = "galaxy_xi"
galaxy_shear_xi = "galaxy_shear_xi"
galaxy_intrinsic_xi = "galaxy_intrinsic_xi"
planck = "planck"
intrinsic_alignment_field = "intrinsic_alignment_field"
fgas_parameters = "fgas_parameters"
shear_calibration = "shear_calibration"
supernova_params = "supernova_params"
bias_field = "bias_field"
galaxy_luminosity_function = "galaxy_luminosity_function"
distances = "distances"
de_equation_of_state = "de_equation_of_state"
growth_parameters = "growth_parameters"
test_parameters = "test_parameters"
mass_function = "mass_function"
evs = "evs"
clusters = "clusters"

